// HttpServer.java - A multithreaded HTTP server.
//
// Copyright (C) 1999,2000  Smart Software Consulting
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Smart Software Consulting
// 1688 Silverwood Court
// Danville, CA  94526-3079
// USA
//
// http://www.smartsc.com
//

package com.smartsc.http;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintWriter;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;

import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.servlet.RequestDispatcher;
import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;

import com.smartsc.util.AbsoluteFile;
import com.smartsc.util.Mailer;
import com.smartsc.util.Properties;

public class
HttpServer
implements ServletContext, HttpStatusCodes, HttpReasonPhrases, Runnable
{
	public
	HttpServer( Properties props)
	throws IOException
	{
		this.props = props;

		// Get verbosity from properties
		boolean verbose = props.getBooleanProperty( PROP_VERBOSE);

		openLog();
		openTransferLog();

		// Log startup message
		banner( verbose);

		// Setup handler counter
		maxHandlers = props.getIntProperty(
			PROP_MAX_HANDLERS, DEFAULT_MAX_HANDLERS);
		handlerCount = 0;

		int backlog = props.getIntProperty( PROP_BACKLOG, -1);
		int port = props.getIntProperty( PROP_PORT, DEFAULT_PORT);
		if( port == 0)
			port = DEFAULT_PORT;

		if( backlog > 0)
		{
			serverSocket = new ServerSocket( port, backlog);
		}
		else
		{
			serverSocket = new ServerSocket( port);
		}

		StringBuffer sb = new StringBuffer(
			InetAddress.getLocalHost().getHostName());
		sb.append( ":");
		sb.append( port);
		serverName = sb.toString();

		String servletPropFile = props.getProperty(
			PROP_SERVLET_PROPS, DEFAULT_SERVLET_PROPS);

		// Load servlets properties file
		try
		{
			File f = AbsoluteFile.create( servletPropFile);
			servletProperties.load( new FileInputStream( f));
		}
		// Ignore exceptions
		catch( IOException ie) {}

		// Preload servlets
		preloadServlets();
	}

	public
	void
	run()
	{
		Socket clientSocket;

		while( true)
		{
			try
			{
				// If we have max number of handlers
				if( handlerCount >= maxHandlers)
				{
					// Wail until some finish
					synchronized( this)
					{
						while( handlerCount >= maxHandlers)
						{
							try
							{
								wait();
								///* TINI Omit
								// (wait is only a stub in TINI Beta 2.1)
								Thread.sleep( 100);
								//*/
							}
							catch( InterruptedException ie) {}
						}
					}
				}

				clientSocket = serverSocket.accept();
				clientSocket.setSoTimeout(
					1000 * props.getIntProperty( PROP_REQUEST_TIMEOUT));
			}
			catch( IOException ioe)
			{
				log( "Exception in accept()", ioe);
				continue;
			}

			try
			{
				HttpServerHandler handler = createHandler( this, clientSocket);
				new Thread( handler).start();
			}
			catch( HttpException he)
			{
				log( he);
				try
				{
					HttpResponse.sendError(
						clientSocket.getOutputStream(),
						he.getStatusCode(),
						he.getReasonPhrase());
				}
				catch( IOException ioe) {}
			}
			catch( Throwable t)
			{
				log( t);
				try
				{
					HttpResponse.sendError(
						clientSocket.getOutputStream(),
						SC_INTERNAL_SERVER_ERROR,
						RP_INTERNAL_SERVER_ERROR);
				}
				catch( IOException ioe) {}
			}
		}
	}

	protected
	void
	openLog()
	{
		String logFileName = props.getProperty( PROP_LOG_FILE);

		if( logFileName != null)
		{
			try
			{
				logFile = AbsoluteFile.create( logFileName);
				logFileName = logFile.getAbsolutePath();
				logWriter = new PrintWriter(
					new FileOutputStream( logFileName, true), true);
			}
			catch( IOException ioe)
			{
				System.err.println(
					"Could not open log file '" +
					logFileName + "' for writing.");
				logWriter = new PrintWriter( System.out, true);
				logFile = null;
			}
		}
		else
		{
			logWriter = new PrintWriter( System.out, true);
			logFile = null;
		}
	}

	protected
	void
	openTransferLog()
	{
		String logFileName = null;
		if( logFile != null)
		{
			logFileName = logFile.getAbsolutePath();
		}

		String transferLogName = props.getProperty( PROP_TRANSFER_LOG);
		if( transferLogName != null)
		{
			transferLogFile = AbsoluteFile.create( transferLogName);
			transferLogName = transferLogFile.getAbsolutePath();
		}

		// Set transferLogWriter = logWriter if any one of these three
		// conditions is true...
		// 1) server.transferLog is not defined
		// 2) server.transferLog is the same as server.logFile
		// 3) server.transferLog is "-" and logFile is not defined
		if( transferLogName == null	|| transferLogName.equals( logFileName)
		||( transferLogName.equals( "-") && transferLogName == null) )
		{
			transferLogWriter = logWriter;
			transferLogFile = logFile;
		}
		else if( transferLogName.equals( "-"))
		{
			transferLogWriter = new PrintWriter( System.out, true);
			transferLogFile = null;
		}
		else
		{
			try
			{
				transferLogWriter = new PrintWriter(
					new FileOutputStream( transferLogName, true), true);
			}
			catch( IOException ioe)
			{
				System.err.println(
					"Could not open transfer log file '" +
					transferLogName + "' for writing.");
				transferLogWriter = logWriter;
				transferLogFile = logFile;
			}
		}
	}

	public
	void
	logTransfer( String msg)
	{
		if( transferLogWriter != null)
		{
			synchronized( transferLogWriter)
			{
				transferLogWriter.println( msg);
			}
		}
	}

	protected
	HttpServerHandler
	createHandler( HttpServer server, Socket clientSocket)
	throws HttpException
	{
		return new HttpServerHandler( server, clientSocket);
	}

	protected synchronized
	void
	addHandler()
	{
		++handlerCount;
	}

	protected synchronized
	void
	removeHandler()
	{
		--handlerCount;
		if( (handlerCount == 0 && getTimeSinceLastGC() > MIN_GC_TIME))
		{
			gc();
		}
		notifyAll();
	}

	protected synchronized
	long
	getTimeSinceLastGC()
	{
		return System.currentTimeMillis() - lastGCTime;
	}

	protected synchronized
	void
	gc()
	{
		// Check log file sizes
		if( logFile != null)
		{
			synchronized( logWriter)
			{
				if( isLogOversized( logFile))
				{
					logWriter.close();
					try
					{
						mailLog( logFile);
						logFile.delete();
					}
					catch( IOException ioe) {}
					openLog();
				}
			}
		}

		if( transferLogFile != null && transferLogFile != logFile)
		{
			synchronized( transferLogWriter)
			{
				if( isLogOversized( transferLogFile))
				{
					transferLogWriter.close();
					try
					{
						mailLog( transferLogFile);
						transferLogFile.delete();
					}
					catch( IOException ioe) {}
					openTransferLog();
				}
			}
		}

		System.gc();
		lastGCTime = System.currentTimeMillis();
	}

	protected void preloadServlets()
	{
		Enumeration e = servletProperties.keys();
		while( e.hasMoreElements())
		{
			String propName = (String)e.nextElement();
			if( propName.endsWith( PRELOAD))
			{
				String propValue = getServletProperty( propName, "false");

				boolean preload = Boolean.valueOf( propValue).booleanValue();
				if( preload)
				{
					String servletClassName = propName.substring( 0,
						propName.length() - PRELOAD.length());

					getServletFromClassName( servletClassName);
				}
			}
		}
	}

	protected Servlet getServletFromPath( String urlpath)
	{
		String servletPrefix = getServletPrefix();
		if( urlpath == null
		||  !urlpath.startsWith( servletPrefix))
		{
			return null;
		}

		// Find servlet name
		int start = servletPrefix.length();
		int slash = urlpath.indexOf( '/', start);
		int hash  = urlpath.indexOf( '#', start);
		int query = urlpath.indexOf( '?', start);

		int end = slash;
		if( end == -1 || (-1 < hash  && hash  < end)) end = hash;
		if( end == -1 || (-1 < query && query < end)) end = query;

		String servletName;
		if( end == -1)
			servletName = urlpath.substring( start);
		else
			servletName = urlpath.substring( start, end);

		return getServletFromName( servletName);
	}


	protected Servlet getServletFromName( String servletName)
	{
		String servletClassName = getServletProperty(
			servletName + CODE, servletName);

		return getServletFromClassName( servletClassName);
	}

	protected Servlet getServletFromClassName( String servletClassName)
	{
		Servlet servlet = null;

		synchronized( servlets)
		{
			// It may already be loaded
			servlet = (Servlet)servlets.get( servletClassName);
			if( servlet != null)
			{
				return servlet;
			}

			Class servletClass;
			try { servletClass = Class.forName( servletClassName); }
			catch( ClassNotFoundException cnfe)
			{
				log( "Class '" + servletClassName + "' not found.");
				return null;
			}

			// If not a servlet
			if( !Servlet.class.isAssignableFrom( servletClass))
			{
				log( servletClassName + " is not a Servlet.");
				return null;
			}

			try { servlet = (Servlet)servletClass.newInstance(); }
			catch( Exception e)
			{
				log( "Could not instantiate Servlet " + servletClassName);
				return null;
			}

			ServletConfig servletConfig =
				new com.smartsc.http.ServletConfig( this,
					getServletProperty(
						servletClassName + INIT_ARGS ) );

			try { servlet.init( servletConfig); }
			catch( Exception e)
			{
				log( "Error initializing Servlet " + servletClassName);
				log( e);
				return null;
			}

			addServlet( servletClassName, servlet);
		}

		return servlet;
	}

	public static
	String
	getContentTypeFor( String fileName)
	{
		String type = "text/plain";

		int lastDot = fileName.lastIndexOf( '.');
		if( lastDot != -1)
		{
			String ext = fileName.substring( lastDot);
			String t = (String)extToMimeType.get( ext);
			if( t != null) type = t;
		}

		return type;
	}

	protected
	boolean
	isLogOversized( File log)
	{
		boolean oversized = false;

		if( isMailingOversizedLogs() && log != null
		&&  log.length() > props.getIntProperty( PROP_MAX_LOG_SIZE) )
		{
			oversized = true;
		}

		return oversized;
	}

	protected
	void
	mailLog( File log)
	throws IOException
	{
		if( mailHeaders == null)
		{
			mailHeaders = new Hashtable();
			mailHeaders.put(
				Mailer.HEADER_TO, props.get( PROP_MAIL_TO));
			mailHeaders.put(
				Mailer.HEADER_FROM, props.get( PROP_MAIL_FROM));
		}
		mailHeaders.put(
			Mailer.HEADER_DATE, new Date().toString());
		mailHeaders.put(
			Mailer.HEADER_SUBJECT, serverName + " " + log.getName());
		try
		{
			// Mail log
			Mailer.send( mailHeaders, log);
		}
		// If bad address
		catch( MalformedURLException mfue)
		{
			// Don't send log anymore
			props.remove( PROP_MAIL_TO);
			props.remove( PROP_MAX_LOG_SIZE);
			throw mfue;
		}
	}

	public
	void
	banner( boolean verbose)
	{
		synchronized( logWriter)
		{
			log( SERVER_INFO);
			log( COPYRIGHT);
			logWriter.println();
			if( verbose)
				props.list( logWriter);
		}
	}

	public static
	void
	help()
	{
		System.out.println( SERVER_INFO);
		System.out.println( COPYRIGHT);
		System.out.println();
		System.out.println( "HttpServer [-(h|v)] [properties_file]");
	}

	public static
	Properties
	getProperties( String[] args)
	{
		boolean help = false;
		boolean verbose = false;

		String propertyFile = null;

		Properties props = new Properties( defaultProps);

		if( args.length > 2)
		{
			help = true;
		}
		else if( args.length == 1)
		{
			if( "-h".equals( args[0])) help = true;
			else if( "-v".equals( args[0])) verbose = true;
			else propertyFile = args[0];
		}
		else if( args.length == 2)
		{
			propertyFile = args[1];
			if( "-v".equals( args[0])) verbose = true;
			else help = true;
		}

		if( help)
		{
			help();
			return null;
		}

		if( propertyFile != null)
		{
			try
			{
				File f = AbsoluteFile.create( propertyFile);
				props.load( new FileInputStream( f));
			}
			catch( IOException ioe)
			{
				System.err.print( "Error reading property file '");
				System.err.print( propertyFile);
				System.err.println( "'.  Using defaults.");
			}
		}

		if( verbose)
		{
			// put() works with both JDK 1.1 and 1.2.
			// setProperty() works with JDK 1.2 only.
			props.put( PROP_VERBOSE, "true");
		}

		// Reconcile path-like properties
		reconcilePathProperty( props, PROP_MIME_TYPES_FILE);
		reconcilePathProperty( props, PROP_DOC_ROOT);
		reconcilePathProperty( props, PROP_SERVLET_PROPS);

		// Load MIME types
		String mimeTypesFilename = props.getProperty( PROP_MIME_TYPES_FILE);
		if( mimeTypesFilename != null)
		{
			try
			{
				extToMimeType.load( new FileInputStream( mimeTypesFilename));
			}
			catch( IOException ioe)
			{
				System.err.print( "Error reading MIME types file '");
				System.err.print( mimeTypesFilename);
				System.err.println( "'.  Using defaults.");
			}
		}

		return props;
	}

	public static
	void
	reconcilePathProperty( Properties props, String pathPropName)
	{
		String pathPropValue = props.getProperty( pathPropName);
		if( pathPropValue == null) return;
		File f = AbsoluteFile.create( pathPropValue);
		pathPropValue = f.getAbsolutePath();
		props.put( pathPropName, pathPropValue);
	}

	public static
	void
	main( String[] args)
	{
		Properties props = getProperties( args);

		if( props != null)
		{
			try
			{
				HttpServer server = new HttpServer( props);
				new Thread( server).start();
			}
			catch( IOException ioe)
			{
				System.err.println( "Could not start server.  (" +
					ioe.getMessage() + ")" );
			}
		}
	}

	private long lastGCTime = System.currentTimeMillis();

	// Server non-properties related constants
	public static final String SERVER_INFO             = "SSC Web Server/0.12";
	public static final String COPYRIGHT =
		"Copyright (C) 1999,2000  Smart Software Consulting.";
	public static final String HTTP_VERSION            = "HTTP/1.1";
	public static final int JSDK_MAJOR_VERSION         = 2;
	public static final int JSDK_MINOR_VERSION         = 1;
	public static final long MIN_GC_TIME               = 30 * 1000;

	// Server properties related constants
	public static final int     DEFAULT_BUFFER_SIZE          = 512;
	public static final String  DEFAULT_DOC_ROOT             = "/docs";
	public static final int     DEFAULT_PORT                 = 80;
	public static final String  DEFAULT_INDEX_FILE           = "index.html";
	public static final int     DEFAULT_MAX_HANDLERS         = 5;
	public static final int     DEFAULT_REQUEST_TIMEOUT      = 10;
	public static final boolean DEFAULT_STACK_TRACE          = false;

	// Servlet constants
	public static final String DEFAULT_SERVLET_PREFIX  = "/servlet/";
//	public static final String DEFAULT_SERVLET_PROPS   = "servlets.properties"; // TINI Omit
	public static final String DEFAULT_SERVLET_PROPS   = "/etc/servlets.props"; // TINI Include

	// Session constants
	public static final int    SESSION_TYPE_COOKIE     = 0;
	public static final int    SESSION_TYPE_URL        = 1;
	public static final int    SESSION_TYPE_NONE       = 2;
	public static final String SESSION_TYPES[]         = { "Cookie", "URL" };
	public static final String DEFAULT_SESSION_NAME    = "sscSessionId";
	public static final int    DEFAULT_SESSION_TIMEOUT = 60*60;  // 1 hour
	public static final String DEFAULT_SESSION_TYPE    = SESSION_TYPES[0];

	// Server-related property names
	public static final String PROP_BACKLOG         = "server.backlog";
	public static final String PROP_BUFFER_SIZE     = "server.bufferSize";
	public static final String PROP_DOC_ROOT        = "server.docRoot";
	public static final String PROP_INDEX_FILE      = "server.indexFile";
	public static final String PROP_LOG_FILE        = "server.logFile";
	public static final String PROP_MAIL_FROM       = "server.mail.from";
	public static final String PROP_MAIL_TO         = "server.mail.to";
	public static final String PROP_MAX_HANDLERS    = "server.maxHandlers";
	public static final String PROP_MAX_LOG_SIZE    = "server.maxLogSize";
	public static final String PROP_MIME_TYPES_FILE = "server.mimeTypesFile";
	public static final String PROP_PORT            = "server.port";
	public static final String PROP_REQUEST_TIMEOUT = "server.requestTimeout";
	public static final String PROP_STACK_TRACE     = "server.stackTrace";
	public static final String PROP_TRANSFER_LOG    = "server.transferLog";
	public static final String PROP_VERBOSE         = "server.verbose";

	// Servlet-related property names
	public static final String PROP_SERVLET_PREFIX  = "servlet.prefix";
	public static final String PROP_SERVLET_PROPS   = "servlet.propFile";

	// Session-related property names
	public static final String PROP_SESSION_NAME    = "session.name";
	public static final String PROP_SESSION_TIMEOUT = "session.timeout";
	public static final String PROP_SESSION_TYPE    = "session.type";

	protected Properties props;
	protected static Properties defaultProps = new Properties();
	static
	{
		// Server default properties
		defaultProps.put(
			PROP_BUFFER_SIZE, String.valueOf( DEFAULT_BUFFER_SIZE));
		defaultProps.put(
			PROP_DOC_ROOT, DEFAULT_DOC_ROOT);
		defaultProps.put(
			PROP_INDEX_FILE, DEFAULT_INDEX_FILE);
		defaultProps.put(
			PROP_MAX_HANDLERS, String.valueOf( DEFAULT_MAX_HANDLERS));
		defaultProps.put(
			PROP_REQUEST_TIMEOUT, String.valueOf( DEFAULT_REQUEST_TIMEOUT));
		defaultProps.put(
			PROP_STACK_TRACE, String.valueOf( DEFAULT_STACK_TRACE));
		defaultProps.put(
			PROP_PORT, String.valueOf( DEFAULT_PORT));
		// Servlet default properties
		defaultProps.put(
			PROP_SERVLET_PREFIX, DEFAULT_SERVLET_PREFIX);
		defaultProps.put(
			PROP_SERVLET_PROPS, DEFAULT_SERVLET_PROPS);
		// Session default properties
		defaultProps.put(
			PROP_SESSION_TYPE, String.valueOf( DEFAULT_SESSION_TYPE));
		defaultProps.put(
			PROP_SESSION_NAME, DEFAULT_SESSION_NAME);
		defaultProps.put(
			PROP_SESSION_TIMEOUT, String.valueOf( DEFAULT_SESSION_TIMEOUT));
	}

	protected int getBufferSize()
	{
		return props.getIntProperty( PROP_BUFFER_SIZE);
	}

	protected String getDocRoot()
	{
		return props.getProperty( PROP_DOC_ROOT);
	}

	protected String getIndexFilename()
	{
		return props.getProperty( PROP_INDEX_FILE);
	}

	protected boolean isMailingOversizedLogs()
	{
		return (props.getProperty( PROP_MAIL_TO) != null)
			&& (props.getIntProperty( PROP_MAX_LOG_SIZE) > 0);
	}

	protected boolean isStackTraceEnabled()
	{
		return props.getBooleanProperty( PROP_STACK_TRACE);
	}

	protected String getServletPrefix()
	{
		return props.getProperty( PROP_SERVLET_PREFIX);
	}

	protected int getSessionType()
	{
		String sessionTypeName = props.getProperty( PROP_SESSION_TYPE);
		for( int i = 0; i < SESSION_TYPES.length; ++i)
		{
			if( SESSION_TYPES[i].equalsIgnoreCase( sessionTypeName))
				return i;
		}

		return SESSION_TYPE_NONE;
	}

	protected String getSessionName()
	{
		return props.getProperty( PROP_SESSION_NAME);
	}

	protected int getSessionTimeout()
	{
		return props.getIntProperty( PROP_SESSION_TIMEOUT);
	}

	private File logFile;
	private PrintWriter logWriter;
	private File transferLogFile;
	private PrintWriter transferLogWriter;
	private Hashtable mailHeaders;

	// TODO Support https?
	public static final String DEFAULT_SCHEME = "http";

	private ServerSocket serverSocket;
	private String serverName;

	private int handlerCount;
	private int maxHandlers  = DEFAULT_MAX_HANDLERS;

	// String attributeName --> String attributeValue
	private Hashtable attributes = new Hashtable();

	// String servletClassName --> Servlet servlet
	private Hashtable servlets = new Hashtable();
	protected Enumeration getServletClassNames()
	{
		return servlets.keys();
	}
	protected void addServlet( String servletClassName, Servlet servlet)
	{
		servlets.put( servletClassName, servlet);
	}
	protected void removeServlet( String servletClassName)
	{
		servlets.remove( servletClassName);
	}

	// servlets.properties stuff
	public static final String CODE      = ".code";
	public static final String INIT_ARGS = ".initArgs";
	public static final String PRELOAD   = ".preload";
	private final Properties servletProperties = new Properties();
	protected String getServletProperty( String key)
	{
		return servletProperties.getProperty( key);
	}
	protected String getServletProperty( String key, String defaultValue)
	{
		String s = servletProperties.getProperty( key);
		if( s == null) s = defaultValue;
		return s;
	}

	// Use java.util.Properties instead of com.smartsc.util.Properties
	// because we don't need the fancy getter methods of the latter.
	private static java.util.Properties
		extToMimeType = new java.util.Properties();
	static
	{
		extToMimeType.put( ".class", "application/octet-stream");
		extToMimeType.put( ".css",   "text/css");
		extToMimeType.put( ".htm",   "text/html");
		extToMimeType.put( ".html",  "text/html");
		extToMimeType.put( ".gif",   "image/gif");
		extToMimeType.put( ".gz",    "application/octet-stream");
		extToMimeType.put( ".jar",   "application/octet-stream");
		extToMimeType.put( ".jpg",   "image/jpeg");
		extToMimeType.put( ".jpeg",  "image/jpeg");
		extToMimeType.put( ".js",    "application/x-javascript");
		extToMimeType.put( ".tini",  "application/octet-stream");
		extToMimeType.put( ".zip",   "application/octet-stream");
	}

	// From javax.servlet.ServletContext
	public Object getAttribute(String name)
	{
		return attributes.get( name);
	}
	public Enumeration getAttributeNames()
	{
		return attributes.keys();
	}
	public ServletContext getContext(String uripath)
	{
		return null;
	}
	public int getMajorVersion()
	{
		return JSDK_MAJOR_VERSION;
	}
	public String getMimeType(String file)
	{
		return getContentTypeFor( file);
	}
	public int getMinorVersion()
	{
		return JSDK_MINOR_VERSION;
	}
	public String getRealPath(String path)
	{
		StringBuffer realPath = new StringBuffer( getDocRoot());

		if( path == null)
			return null;

		// TODO path aliases
		// TODO prevent .. from going above docRoot
		if( !path.startsWith( "/"))
		{
			realPath.append( "/");
		}
		realPath.append( path);
		return realPath.toString();
	}
	public RequestDispatcher getRequestDispatcher(String urlpath)
	{
		Servlet servlet = getServletFromPath( urlpath);
		if( servlet == null)
			return null;

		RequestDispatcher rd = new
			com.smartsc.http.RequestDispatcher( servlet, urlpath);

		return rd;
	}

	public URL getResource(String path)
	throws MalformedURLException
	{
		// TODO Support URL resources
		return null;
	}

	public InputStream getResourceAsStream(String path)
	{
		// Build filename
		String filename = getRealPath( path);

		// Construct new file object
		File file = new File( filename);

		// If file does not exist
		// or can not be read
		if( !file.exists()
		||  !file.canRead() )
		{
			return null;
		}

		// If directory
		if( file.isDirectory())
		{
			file = new File( file, getIndexFilename());
		}

		// Get length
		long length = file.length();

		// If file does not exist
		// or can not be read
		// or has zero length
		if( !file.exists()
		||  !file.canRead()
		||  length == 0 )
		{
			return null;
		}

		FileInputStream in = null;
		try { in = new FileInputStream(file); }
		catch( FileNotFoundException fnfe) {}

		return in;
	}
	public String getServerInfo()
	{
		return SERVER_INFO;
	}
	/** @deprecated */
	public Servlet getServlet(String name)
	{
		return null;
	}
	/** @deprecated */
	public Enumeration getServletNames()
	{
		return new Vector(0).elements();
	}
	/** @deprecated */
	public Enumeration getServlets()
	{
		return new Vector(0).elements();
	}
	/** @deprecated */
	public void log( Exception exception, String msg)
	{
		log( msg, exception);
	}
	public void log( String msg)
	{
		// TODO Add timestamp
		if( msg != null && msg.length() > 0)
		{
			synchronized( logWriter)
			{
				logWriter.println( msg);
			}
		}
	}
	public void log( String msg, Throwable throwable)
	{
		synchronized( logWriter)
		{
			log( msg);
			if( isStackTraceEnabled())
			{
				throwable.printStackTrace( logWriter);
			}
		}
	}
	public void log( Throwable t)
	{
		log( t.getClass().getName() + ": " + t.getMessage(), t);
	}
	public void removeAttribute(String name)
	{
		attributes.remove( name);
	}
	public void setAttribute(String name, Object o)
	{
		attributes.put( name, o);
	}
}
