// HttpServerHandler.java - Request handler for HttpServer.
//
// Copyright (C) 1999,2000  Smart Software Consulting
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Smart Software Consulting
// 1688 Silverwood Court
// Danville, CA  94526-3079
// USA
//
// http://www.smartsc.com
//

package com.smartsc.http;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

import java.net.Socket;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.SingleThreadModel;

public class
HttpServerHandler
implements HttpStatusCodes, HttpReasonPhrases, HttpMethods, Runnable
{
	// Protected contructor
	protected
	HttpServerHandler( HttpServer server, Socket socket)
	throws HttpException
	{
		this.server = server;
		this.socket = socket;
	}

	protected
	boolean
	init()
	{
		boolean initOk = true;

		try
		{
			req = new HttpRequest( server, socket);
			res = new HttpResponse( server, req, socket);
		}
		catch( Exception e)
		{
			server.log( e);
			if( socket != null)
			{
				try { socket.close(); }
				catch( IOException ioe) {}
			}
			initOk = false;
		}

		return initOk;
	}

	public final
	void
	run()
	{
		// TODO handle "Keep-Alive"
		boolean initOk = init();

		if( initOk)
		{
			try
			{
				server.addHandler();
				req.parseRequest();
				service( req, res);
			}
			catch( HttpException he)
			{
				server.log( he);
				res.sendError(
					he.getStatusCode(),
					he.getReasonPhrase());
			}
			catch( Exception e)
			{
				server.log( e);
				res.sendError(
					SC_INTERNAL_SERVER_ERROR,
					RP_INTERNAL_SERVER_ERROR);
			}
			finally
			{
				closeConnection();
				server.removeHandler();
			}
		}
	}

	public
	void
	service( HttpRequest req, HttpResponse res)
	throws HttpException
	{
		// Get url path
		String urlpath = req.getRequestPath();

		// If servlet
		if( urlpath.startsWith( server.getServletPrefix()) )
		{
			//Invoke servlet
			Servlet servlet = server.getServletFromPath( urlpath);

			if( servlet == null)
				throw new HttpException( SC_NOT_FOUND,
					RP_NOT_FOUND + " (" + urlpath + ")");

			invokeServlet( servlet, req, res);

			// Done
			return;
		}

		String method = req.getMethod();
		if( GET.equals( method))
		{
			doGet( req, res);
		}
		// TODO support other methods
		else
		{
			throw new HttpException(
				SC_METHOD_NOT_ALLOWED, RP_METHOD_NOT_ALLOWED);
		}
	}

	protected
	void
	invokeServlet( Servlet servlet, HttpRequest req, HttpResponse res)
	{
		try
		{
			if( servlet instanceof SingleThreadModel)
			{
				synchronized( servlet)
				{
					servlet.service( req, res);
				}
			}
			else
			{
				servlet.service( req, res);
			}
		}
		catch( Exception e)
		{
			server.log( e);
			try
			{
				res.sendError( SC_INTERNAL_SERVER_ERROR, e.getMessage());
			}
			catch( Exception e1) {}
		}
	}

	protected
	void
	doGet( HttpRequest req, HttpResponse res)
	throws HttpException
	{
		// Get request URI (ignore any query string)
		String path = req.getRequestURI();

		// Build filename to determine MIME type
		String filename = server.getRealPath( path);
		File file = new File( filename);

		// If directory
		if( file.isDirectory())
		{
			// Use index filename
			filename = server.getIndexFilename();
			file = new File( file, filename);
		}

		// If file is not under doc root
		if( !file.getAbsolutePath().startsWith( server.getDocRoot()) )
		{
			throw new HttpException( SC_FORBIDDEN, RP_FORBIDDEN);
		}

		// If file does not exist
		if( !file.exists())
		{
			throw new HttpException( SC_NOT_FOUND,
					RP_NOT_FOUND + " (" + path + ")");
		}

		// Set Last Modified header, if not already set
		long lastModified = file.lastModified();
		if( !res.containsHeader( HttpHeaders.LAST_MODIFIED))
		{
			res.setDateHeader( HttpHeaders.LAST_MODIFIED, lastModified);
		}



		long ifModifiedSince =
			req.getDateHeader( HttpHeaders.IF_MODIFIED_SINCE);

		if (ifModifiedSince >= 0)
		{
			long now = System.currentTimeMillis();

			if( lastModified <= ifModifiedSince
			&&  ifModifiedSince <= now )
			{
				res.sendError(HttpStatusCodes.SC_NOT_MODIFIED);
			}
		}

		// Get input stream for request path
		InputStream is = new BufferedInputStream(
			server.getResourceAsStream( path),
			server.getBufferSize() );

		// Set content type
		try { res.setContentType( server.getContentTypeFor( filename)); }
		catch( IllegalStateException ise) {}

		try { res.setContentLength( is.available()); }
		catch( IOException ioe)
		{
			throw new HttpException(
				SC_INTERNAL_SERVER_ERROR, RP_INTERNAL_SERVER_ERROR);
		}

		OutputStream os = null;
		try
		{
			os = res.getOutputStream();
			byte[] buf = new byte[server.getBufferSize()];

			int bytesRead = is.read( buf);
			while( bytesRead != -1 )
			{
				os.write( buf, 0, bytesRead);
				bytesRead = is.read( buf);
			}
		}
		catch( IOException ioe)
		{
			server.log( ioe);
		}
		finally
		{
			try
			{
				is.close();
				os.flush();
				os.close();
			}
			catch( IOException e) {}
		}
	}

	protected
	void
	closeConnection()
	{
		if( res != null)
		{
			try { res.close(); }
			catch( IOException ioe) {}
		}

		if( req != null)
		{
			try { req.close(); }
			catch( IOException ioe) {}
		}

		if( socket != null)
		{
			try { socket.close(); }
			catch( IOException ioe) {}
		}
	}
	private HttpServer server;
	private Socket socket;
	private HttpRequest req;
	private HttpResponse res;
}
